/*----------------------------------------------------------------------------
 *
 * qgram.c
 *
 * N-Gram distance
 *
 * FIXME the function is the same as block (L1 distance); the only
 * FIXME difference is that it uses tokenizeByGram. Consider doing
 * FIXME _block() function and using it for both functions.
 *
 *
 * Copyright (c) 2008-2009, Euler Taveira de Oliveira
 *
 *----------------------------------------------------------------------------
 */

#include "similarity.h"
#include "tokenizer.h"

/*
 */

int	qgram_tokenizer = PGS_UNIT_GRAM;
float4	qgram_threshold = 0.7f;
bool	qgram_is_normalized = true;

PG_FUNCTION_INFO_V1(qgram);

Datum
qgram(PG_FUNCTION_ARGS)
{
	char		*a, *b;
	TokenList	*s, *t, *u;
	Token		*p, *q, *r;
	float		maxmatching;
	float		difference;
	float4		res;

	a = DatumGetPointer(DirectFunctionCall1(textout, PointerGetDatum(PG_GETARG_TEXT_P(0))));
	b = DatumGetPointer(DirectFunctionCall1(textout, PointerGetDatum(PG_GETARG_TEXT_P(1))));

	if (strlen(a) > PGS_MAX_STR_LEN || strlen(b) > PGS_MAX_STR_LEN)
		ereport(ERROR,
				(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
				errmsg("argument exceeds the maximum length of %d bytes",
					PGS_MAX_STR_LEN)));

	/* lists */
	s = initTokenList(0);
	t = initTokenList(0);
	/* set list */
	u = initTokenList(1);

	switch (qgram_tokenizer)
	{
		case PGS_UNIT_WORD:
			tokenizeBySpace(s, a);
			tokenizeBySpace(t, b);
			/* all tokens in a set */
			tokenizeBySpace(u, a);
			tokenizeBySpace(u, b);
			break;
		case PGS_UNIT_GRAM:
		default:
			tokenizeByGram(s, a);
			tokenizeByGram(t, b);
			/* all tokens in a set */
			tokenizeByGram(u, a);
			tokenizeByGram(u, b);
			break;
		case PGS_UNIT_ALNUM:	/* default */
			tokenizeByNonAlnum(s, a);
			tokenizeByNonAlnum(t, b);
			/* all tokens in a set */
			tokenizeByNonAlnum(u, a);
			tokenizeByNonAlnum(u, b);
			break;
	}

	elog(DEBUG3, "Token List A");
	printToken(s);
	elog(DEBUG3, "Token List B");
	printToken(t);
	elog(DEBUG3, "All Token List");
	printToken(u);

	maxmatching = s->size + t->size;
	difference = 0.0;

	p = u->head;
	while (p != NULL)
	{
		int		acnt = 0;
		int		bcnt = 0;

		q = s->head;
		while (q != NULL)
		{
			elog(DEBUG4, "p: %s; q: %s", p->data, q->data);
			if (strcmp(p->data, q->data) == 0)
				acnt++;
			q = q->next;
		}

		r = t->head;
		while (r != NULL)
		{
			elog(DEBUG4, "p: %s; r: %s", p->data, r->data);
			if (strcmp(p->data, r->data) == 0)
				bcnt++;
			r = r->next;
		}

		if (acnt > bcnt)
			difference += (acnt - bcnt);
		else
			difference += (bcnt - acnt);

		elog(DEBUG2,
			"\"%s\" => acnt(%d); bcnt(%d); difference(%.2f)",
					p->data, acnt, bcnt, difference);

		p = p->next;
	}

	elog(DEBUG1, "is normalized: %d", qgram_is_normalized);
	elog(DEBUG1, "maximum matching: %.2f", maxmatching);
	elog(DEBUG1, "difference: %.2f", difference);

	destroyTokenList(s);
	destroyTokenList(t);
	destroyTokenList(u);

	if (qgram_is_normalized)
	{
		if (maxmatching == 0.0)
			res = 0.0;
		else
			res = (maxmatching - difference) / maxmatching;
	}
	else
	{
		res = difference;
	}

	PG_RETURN_FLOAT4(res);
}

PG_FUNCTION_INFO_V1(qgram_op);

Datum qgram_op(PG_FUNCTION_ARGS)
{
	float4	res;

	/*
	 * store *_is_normalized value temporarily 'cause
	 * threshold (we're comparing against) is normalized
	 */
	bool	tmp = qgram_is_normalized;
	qgram_is_normalized = true;

	res = DatumGetFloat4(DirectFunctionCall2(
					qgram,
					PG_GETARG_DATUM(0),
					PG_GETARG_DATUM(1)));

	/* we're done; back to the previous value */
	qgram_is_normalized = tmp;

	PG_RETURN_BOOL(res >= qgram_threshold);
}
