#ifndef SIMILARITY_H
#define	SIMILARITY_H

#include "postgres.h"

#include "fmgr.h"
#include "utils/builtins.h"

/* case insensitive ? */
#define		PGS_IGNORE_CASE			1

/* maximum string length */
#define		PGS_MAX_STR_LEN			1024

/*
 * Jaro
 */

/* operation's weight */
#define	PGS_JARO_W1					1.0/3.0
#define	PGS_JARO_W2					1.0/3.0
#define	PGS_JARO_WT					1.0/3.0

/* size of the initial prefix considered */
#define	PGS_JARO_PREFIX_SIZE		4

/* scaling factor */
#define	PGS_JARO_SCALING_FACTOR		0.1

/* minimum score for a string that gets boosted */
#define	PGS_JARO_BOOST_THRESHOLD	0.7

/*
 * Levenshtein
 */
#define		PGS_LEV_MIN_COST		0
#define		PGS_LEV_MAX_COST		1

/*
 * Needleman-Wunch
 */
#define		PGS_NW_GAP_COST			-5

/*
 * Smith-Waterman
 */
#define		PGS_SW_MIN_COST			-2.0
#define		PGS_SW_MAX_COST			1.0
#define		PGS_SW_GAP_COST			0.5

/*
 * Smith-Waterman-Gotoh
 */
#define		PGS_SWG_WINDOW_SIZE		100

/*
 * Soundex
 */
#define		PGS_SOUNDEX_LEN			4
#define		PGS_SOUNDEX_INV_CODE		-1

/*
 * commonly used functions
 */
#define		min2(a, b)			((a < b) ? a : b)
#define		max2(a, b)			((a > b) ? a : b)
#define		min3(a, b, c)		((a < b && a < c) ? a : ((b < c) ? b : c))
#define		max3(a, b, c)		((a > b && a > c) ? a : ((b > c) ? b : c))
#define		max4(a, b, c, d)	((a > b && a > c) ? a : ((b > c && b > d) ? b : ((c > d) ? c : d)))

/*
 * normalized results?
 */
extern bool	block_is_normalized;
extern bool	cosine_is_normalized;
extern bool	dice_is_normalized;
extern bool	euclidean_is_normalized;
extern bool	hamming_is_normalized;
extern bool	jaccard_is_normalized;
extern bool	jaro_is_normalized;
extern bool	jarowinkler_is_normalized;
extern bool	levenshtein_is_normalized;
extern bool	matching_is_normalized;
extern bool	mongeelkan_is_normalized;
extern bool	needlemanwunsch_is_normalized;
extern bool	overlap_is_normalized;
extern bool	qgram_is_normalized;
extern bool	smithwaterman_is_normalized;
extern bool	smithwatermangotoh_is_normalized;

/*
 * how to separate things?
 */
enum
{
	PGS_UNIT_WORD,		/* tokenize by spaces */
	PGS_UNIT_GRAM,		/* tokenize by n-gram */
	PGS_UNIT_ALNUM		/* tokenize by nonalnum characters */
};

/*
 * tokenizers per function
 */
extern int	block_tokenizer;
extern int	cosine_tokenizer;
extern int	dice_tokenizer;
extern int	euclidean_tokenizer;
extern int	jaccard_tokenizer;
extern int	matching_tokenizer;
extern int	mongeelkan_tokenizer;
extern int	overlap_tokenizer;
extern int	qgram_tokenizer;

/*
 * thresholds per function
 */
extern float4	block_threshold;
extern float4	cosine_threshold;
extern float4	dice_threshold;
extern float4	euclidean_threshold;
extern float4	hamming_threshold;
extern float4	jaccard_threshold;
extern float4	jaro_threshold;
extern float4	jarowinkler_threshold;
extern float4	levenshtein_threshold;
extern float4	matching_threshold;
extern float4	mongeelkan_threshold;
extern float4	needlemanwunsch_threshold;
extern float4	overlap_threshold;
extern float4	qgram_threshold;
extern float4	smithwaterman_threshold;
extern float4	smithwatermangotoh_threshold;

/*
 * levenshtein.c
 */
int _lev(char *a, char *b, int icost, int dcost);
int _lev_slow(char *a, char *b, int icost, int dcost);

/*
 * similarity.c
 */
int levcost(char a, char b);
int nwcost(char a, char b);
float swcost(char *a, char *b, int i, int j);
float swggapcost(int i, int j);
float megapcost(char *a, char *b, int i, int j);

/*
 * external function declarations
 */
extern Datum block(PG_FUNCTION_ARGS);
extern Datum block_op(PG_FUNCTION_ARGS);
extern Datum cosine(PG_FUNCTION_ARGS);
extern Datum cosine_op(PG_FUNCTION_ARGS);
extern Datum dice(PG_FUNCTION_ARGS);
extern Datum dice_op(PG_FUNCTION_ARGS);
extern Datum euclidean(PG_FUNCTION_ARGS);
extern Datum euclidean_op(PG_FUNCTION_ARGS);
extern Datum hamming(PG_FUNCTION_ARGS);
extern Datum hamming_op(PG_FUNCTION_ARGS);
extern Datum jaccard(PG_FUNCTION_ARGS);
extern Datum jaccard_op(PG_FUNCTION_ARGS);
extern Datum jaro(PG_FUNCTION_ARGS);
extern Datum jaro_op(PG_FUNCTION_ARGS);
extern Datum jarowinkler(PG_FUNCTION_ARGS);
extern Datum jarowinkler_op(PG_FUNCTION_ARGS);
extern Datum lev(PG_FUNCTION_ARGS);
extern Datum lev_op(PG_FUNCTION_ARGS);
extern Datum levslow(PG_FUNCTION_ARGS);
extern Datum levslow_op(PG_FUNCTION_ARGS);
extern Datum matchingcoefficient(PG_FUNCTION_ARGS);
extern Datum matchingcoefficient_op(PG_FUNCTION_ARGS);
extern Datum mongeelkan(PG_FUNCTION_ARGS);
extern Datum mongeelkan_op(PG_FUNCTION_ARGS);
extern Datum needlemanwunsch(PG_FUNCTION_ARGS);
extern Datum needlemanwunsch_op(PG_FUNCTION_ARGS);
extern Datum overlapcoefficient(PG_FUNCTION_ARGS);
extern Datum overlapcoefficient_op(PG_FUNCTION_ARGS);
extern Datum qgram(PG_FUNCTION_ARGS);
extern Datum qgram_op(PG_FUNCTION_ARGS);
extern Datum smithwaterman(PG_FUNCTION_ARGS);
extern Datum smithwaterman_op(PG_FUNCTION_ARGS);
extern Datum smithwatermangotoh(PG_FUNCTION_ARGS);
extern Datum smithwatermangotoh_op(PG_FUNCTION_ARGS);

extern Datum set_isnormalized(PG_FUNCTION_ARGS);
extern Datum get_isnormalized(PG_FUNCTION_ARGS);

extern Datum set_threshold(PG_FUNCTION_ARGS);
extern Datum get_threshold(PG_FUNCTION_ARGS);

extern Datum set_tokenizer(PG_FUNCTION_ARGS);
extern Datum get_tokenizer(PG_FUNCTION_ARGS);

#endif
